use crate::filter::Filter;

use self::property::Property;
use domo_proto::{identifier::Identifier, packet::Packet};
use std::collections::HashMap;

pub mod property;

/// Simple Node abstraction.
#[derive(Debug, Clone)]
pub struct Node {
    /// Who owns this node?
    pub parent: Option<Identifier>,
    /// Identifier of node
    pub identifier: Identifier,
    /// Child nodes
    children: HashMap<Identifier, Node>,
    /// All properties of the node.
    pub properties: HashMap<String, Property>,
}

impl Node {
    /// Create a new node
    pub fn new(node_id: Identifier) -> Self {
        Node {
            parent: None,
            identifier: node_id,
            children: HashMap::new(),
            properties: HashMap::new(),
        }
    }

    /// Attempts to add child node to self.
    /// Returns a mutable reference to the added node.
    pub fn add_child<'a>(&'a mut self, mut node: Node) -> Option<&'a mut Node> {
        if self.children.contains_key(&node.identifier) {
            return None;
        }
        node.parent = Some(self.identifier.clone());
        self.children.insert(node.identifier, node.clone());
        self.children.get_mut(&node.identifier)
    }

    /// Attempts to remove child from self.
    /// Returns removed node.
    pub fn remove_child(&mut self, identifier: &Identifier) -> Option<Node> {
        self.children.remove(&identifier).map(|mut n| {
            n.parent = None;
            n
        })
    }

    /// Returns if this node has the identifier.
    pub fn has(&self, identifier: &Identifier) -> bool {
        self.children.contains_key(identifier)
    }

    /// Attempts to find given identifier and returns mutable reference to it.
    pub fn find<'a>(&'a mut self, identifier: &Identifier) -> Option<&'a mut Node> {
        for (_, child) in self.children.iter_mut() {
            if let Some(found) = child.find(identifier) {
                return Some(found);
            }
        }
        None
    }
}

impl Filter<Identifier> for Node {
    fn filter(
        &self,
        f: crate::filter::FilterState<Identifier>,
    ) -> crate::filter::FilterState<Identifier> {
        if let Some(i) = f.to_type() {
            if self.has(&i) || i == self.identifier {
                crate::filter::FilterState::Pass(i)
            } else {
                crate::filter::FilterState::Skip(i)
            }
        } else {
            crate::filter::FilterState::Drop
        }
    }
}

impl Filter<Packet> for Node {
    fn filter(&self, f: crate::filter::FilterState<Packet>) -> crate::filter::FilterState<Packet> {
        if let Some(p) = f.to_type() {
            if self.has(&p.dest) || self.identifier == p.dest {
                crate::filter::FilterState::Pass(p)
            } else {
                crate::filter::FilterState::Skip(p)
            }
        } else {
            crate::filter::FilterState::Drop
        }
    }
}
