use domo_proto::packet::Packet;

use crate::filter::{
    filters::function_filter::{FunctionFilter, FunctionFilterFunction},
    Filter, FilterState,
};

use self::endpoint::Endpoint;

pub mod endpoint;

/// The Router will first filter the packet through the filter chain.
/// Then it will filter past all endpoints and send the packet if the endpoint filter passed.
pub struct Router<'a> {
    filters: Vec<Box<dyn Filter<Packet> + 'a>>,
    endpoints: Vec<Box<dyn Endpoint<Packet> + 'a>>,
}

impl<'a> Router<'a> {
    pub fn new() -> Router<'a> {
        Router {
            filters: vec![],
            endpoints: vec![],
        }
    }

    pub fn add_filter(&mut self, filter: impl Filter<Packet> + 'a) {
        self.filters.push(Box::new(filter));
    }

    pub fn add_function_filter(&mut self, function: FunctionFilterFunction) {
        self.filters.push(Box::new(FunctionFilter(function)))
    }

    pub fn add_endpoint(&mut self, endpoint: impl Endpoint<Packet> + 'a) {
        self.endpoints.push(Box::new(endpoint))
    }
}

impl Filter<Packet> for Router<'_> {
    fn filter(&self, f: FilterState<Packet>) -> FilterState<Packet> {
        let mut f = f;
        for filter in &self.filters {
            f = filter.filter(f);
        }
        f
    }
}

impl Endpoint<Packet> for Router<'_> {
    fn handle(&mut self, item: Packet) {
        let item = self.filter(FilterState::Pass(item));
        for endpoint in &mut self.endpoints {
            match endpoint.filter(item.clone()) {
                FilterState::Pass(p) | FilterState::Skip(p) => {
                    let _ = &endpoint.handle(p);
                }
                FilterState::Drop => {}
            }
        }
    }
}
